/*
 * file ResourceSource.java
 * 
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM - you are allowed to copy, modify and 
 * redistribute this file as part of any program that interfaces with 
 * IBM Rational CM API.
 *
 * com.ibm.rational.teamapi.scout.ResourceSource
 *
 * (C) Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */
package com.ibm.rational.teamapi.scout;

import java.lang.reflect.Array;
import java.util.List;

import javax.wvcm.PropertyRequestItem;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyNameList.PropertyName;
import javax.wvcm.PropertyRequestItem.PropertyRequest;

import org.eclipse.swt.custom.BusyIndicator;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.views.properties.IPropertyDescriptor;
import org.eclipse.ui.views.properties.IPropertySheetEntry;
import org.eclipse.ui.views.properties.PropertyDescriptor;

import com.ibm.rational.wvcm.stp.StpLocation;
import com.ibm.rational.wvcm.stp.StpProperty;
import com.ibm.rational.wvcm.stp.StpPropertyException;
import com.ibm.rational.wvcm.stp.StpResource;
import com.ibm.rational.wvcm.stp.StpProperty.MetaPropertyName;


/**
 * An implementation of IPropertySource for displaying the properties of a CM
 * API Resource. ProperyName objects are used for the property identifiers.
 */
public class ResourceSource
    extends DefaultPropertySource
{
    /**
     * Requests all properties from the server and then constructs a property
     * descriptor for each property returned. For efficiently, the descriptors
     * are cached after the properties are read from the resource. The
     * possibility of displaying stale data could be avoided by re-reading the
     * properties each time the descriptors are requested.
     *
     * @return  An array of ResourcePropertyDescriptors, one for each property
     *          of the resource represented by this object.
     */
    public IPropertyDescriptor[] getPropertyDescriptors()
    {
        if (m_descriptors == null) {
            try {
                // At the top level some pseudo-resources are used to represent
                // requests for folder lists. Properties should not be requested
                // from these.
                StpLocation selector = m_resource.stpLocation();
                String      name = selector.getName();
                String      server = "";
                String      repo = selector.getRepo();

                if ((name == null || name.length() == 0)
                        && (repo == null || repo.length() == 0)
                        && (server != null && server.length() > 0)) {
                    return null;
                }

                // This may take a while, so run it off the UI thread and
                // put up a wait cursor.
                final Display  display = Display.getCurrent();
                final Runnable longJob =
                    new Runnable() {
                        boolean m_done = false;

                        public void run()
                        {
                            Thread thread = new Thread(new Runnable() {
                                        public void run()
                                        {
                                            try {
                                                m_resource =
                                                    (StpResource)m_resource
                                                    .doReadProperties(
                                                        WANTED_PROPS);
                                            } catch (WvcmException e) {
                                                e.printStackTrace();
                                            }
                                            m_done = true;
                                        }
                                    });
                            thread.start();

                            while (!m_done) {
                                if (!display.readAndDispatch())
                                    display.sleep();
                            }
                        }
                    };

                BusyIndicator.showWhile(display, longJob);

                // Get a list of all the properties returned in the new proxy.
                StpProperty.List properties =
                    (StpProperty.List)m_resource.getAllProperties();

                m_descriptors = new IPropertyDescriptor[properties.size()];

                for (int i = 0; i < m_descriptors.length; ++i) {
                    Object property = properties.get(i);

                    // The StpProperty.List will contain a PropertyException if
                    // the requested property could not be returned from the
                    // resource; a StpProperty object otherwise. A different
                    // variant of ResourcePropertyDescriptor is used for each
                    // case.
                    if (property instanceof StpProperty)
                        m_descriptors[i] =
                            new ResourcePropertyDescriptor((StpProperty)
                                property,
                                m_resource);
                    else
                        m_descriptors[i] =
                            new ResourcePropertyDescriptor(
                                (StpPropertyException)property,
                                m_resource);
                }
            } catch (Throwable ex) {
                ex.printStackTrace();
            }
        }

        return m_descriptors;
    }

    /**
     * Retrieves the value of a property from the resource proxy. Compound
     * values are wrapped by a type-specific PropertySource for viewing the
     * values.
     *
     * @param  id  The property identifier, which, in this case, is the
     *             PropertyName of the CM API property.
     *
     * @return  Either a PropertySource (for compound values) or the value
     *          itself.
     */
    public Object getPropertyValue(Object id)
    {
        Object val = m_resource.lookupProperty((PropertyName)id);

        return getPropertySource(val);
    }

    /**
     * Constructs a ResourceSource object for displaying the properties of a
     * resource identified by a Resource proxy.
     *
     * @param  resource  A Resource proxy for the resource whose properties are
     *                   to be displayed.
     */
    ResourceSource(StpResource resource)
    {
        m_resource = resource;
    }

    /**
     * The proxy for the resource whose properties are displayed by this source
     */
    protected StpResource m_resource;

    /** The cached property descriptors for the properties read. */
    IPropertyDescriptor[] m_descriptors = null;

    /** The PropertyRequest used to request all properties from the resource */
    static final PropertyRequest WANTED_PROPS =
        new PropertyRequest(
            new PropertyRequestItem[] {StpResource.ALL_PROPERTIES.nest(
                    new PropertyRequestItem[] {StpProperty.VALUE.nest(
                            new PropertyRequestItem[] {
                                StpProperty.NAME,
                                StpProperty.TYPE,
                                StpProperty.VALUE
                            })})});

    /** The value used to flag properties with errors */
    static final String[] ADVANCED_PROPERTY_FLAGS =
        new String[] {IPropertySheetEntry.FILTER_ID_EXPERT};

    /**
     * A PropertyDescriptor for CM API resource properties.
     */
    static class ResourcePropertyDescriptor
        extends PropertyDescriptor
    {
        /**
         * Constructs a PropertyDescriptor for a StpProperty successfully
         * retrieved from a proxy. If the VALUE meta-property of the StpProperty
         * is inaccessible, the property is marked as one for expert viewing
         * only.
         *
         * @param  property  The StpProperty
         * @param  parent  The resource from which the StpProperty came.
         */
        ResourcePropertyDescriptor(
            StpProperty property,
            StpResource parent)
        {
            super(property.getPropertyName(),
                property.getPropertyName().getName());
            m_propertyName = property.getPropertyName();

            try {
                property.getValue();
            } catch (WvcmException ex) {
                m_filter_flags = ADVANCED_PROPERTY_FLAGS;
            }

            try {
                setDescription(property.getType() + " property of "
                    + parent.location().string());
            } catch (WvcmException ex) {
                setDescription("Untyped property of "
                    + parent.location().string());
            }
        }

        /**
         * Constructs a PropertyDescriptor for a PropertyException received
         * while attempting to retrieve a StpProperty from a proxy. The property
         * is marked as one for expert viewing only.
         *
         * @param  property  The PropertyException
         * @param  parent  The resource from which the StpProperty came.
         */
        ResourcePropertyDescriptor(
            StpPropertyException property,
            StpResource          parent)
        {
            super(property.getPropertyName(),
                property.getPropertyName().getName());
            m_propertyName = property.getPropertyName();
            m_filter_flags = ADVANCED_PROPERTY_FLAGS;
            setDescription("inaccessible property of "
                + parent.location().string());
        }

        /**
         * Returns the Namespace of the PropertyName for the property to use as
         * a category for organizing the properties of the resource.
         *
         * @return  A String containing the simple name of the namespace of the
         *          property name.
         */
        public String getCategory()
        {
            String category = m_propertyName.getNamespace();

            if (category == null)
                category = "WVCM";
            else {
                int slash = category.lastIndexOf("/");

                if (slash > 0)
                    category = category.substring(slash + 1);
            }

            return category;
        }

        /**
         * Returns ADVANCED_PROPERTY_FLAGS for properties that could not be read
         * from the resource; null otherwise;
         *
         * @see  org.eclipse.ui.views.properties.PropertyDescriptor#getFilterFlags()
         */
        public String[] getFilterFlags()
        {
            return m_filter_flags;
        }

        /** The PropertyName for the property */
        PropertyName m_propertyName;

        /** The value of getFilterFlags() */
        String[] m_filter_flags = super.getFilterFlags();
    }

    /**
     * A PropertySource for displaying properties whose value is a list.
     */
    static class ListSource
        extends DefaultPropertySource
    {
        /**
         * Constructs a new ListSource for a List value.
         *
         * @param  list  The value that is to be displayed by this
         *               PropertySource
         */
        ListSource(List list)
        {
            m_list = list;
        }

        /**
         * Computes and returns an array of PropertyDescriptors, one descriptor
         * per element of the list. An Integer object representing the ordinal
         * position of the element in the list is used as the property id.
         *
         * @return  An array of PropertyDescriptors for the elements of the
         *          ResourceList value of this property.
         */
        public IPropertyDescriptor[] getPropertyDescriptors()
        {
            IPropertyDescriptor[] result =
                new IPropertyDescriptor[m_list.size()];

            for (int i = 0; i < result.length; ++i) {
                String name = getDisplayName(m_list.get(i), i);

                result[i] = new PropertyDescriptor(new Integer(i), name);
                ((PropertyDescriptor)result[i]).setDescription(m_list.getClass()
                    .getName() + " item");
            }

            return result;
        }

        /**
         * Retrieves the List item identified by a given property identifier.
         *
         * @param  id  An Integer representing the ordinal position of the item
         *             in the list
         *
         * @return  The value at the given index, wrapped in a PropertySource if
         *          the item is a composite object such as a Resource, List, or
         *          Array.
         */
        public Object getPropertyValue(Object id)
        {
            return getPropertySource(m_list.get(((Integer)id).intValue()));
        }

        /** The List that this PropertySource will be displaying */
        List m_list;
    }

    /**
     * A PropertySource for displaying properties whose value is an array.
     */
    static class ArraySource
        extends DefaultPropertySource
    {
        /**
         * Constructs an ArraySource for an array value
         *
         * @param  array  The array to be displayed by the ArraySource.
         */
        ArraySource(Object array)
        {
            m_array = array;
        }

        /**
         * Computes and returns an array of PropertyDescriptors, one descriptor
         * per element of the array. An Integer object representing the ordinal
         * position of the element in the array is used as the property id.
         *
         * @return  An array of PropertyDescriptors for the elements of this
         *          array-valued property.
         */
        public IPropertyDescriptor[] getPropertyDescriptors()
        {
            IPropertyDescriptor[] result =
                new IPropertyDescriptor[Array.getLength(m_array)];

            for (int i = 0; i < result.length; ++i) {
                String name = getDisplayName(Array.get(m_array, i), i);

                result[i] = new PropertyDescriptor(new Integer(i), name);
            }

            return result;
        }

        /**
         * Retrieves the array element identified by a given property
         * identifier.
         *
         * @param  id  An Integer representing the index of the item in the
         *             array
         *
         * @return  The value at the given index, wrapped in a PropertySource if
         *          the item is a composite object such as a Resource, List, or
         *          Array.
         */
        public Object getPropertyValue(Object id)
        {
            return getPropertySource(Array.get(
                        m_array,
                        ((Integer)id).intValue()));
        }

        /** The array of values represented by this PropertySource */
        Object m_array;
    }

    /**
     * A PropertySource for displaying properties whose value is a StpProperty.
     */
    static class PropertySource
        extends DefaultPropertySource
    {
        /**
         * Constructs a PropertySource for a StpProperty.
         *
         * @param  property  The StpProperty object to be displayed by this
         *                   PropertySource
         */
        PropertySource(StpProperty property)
        {
            m_property = property;
        }

        /**
         * Computes and returns an array of PropertyDescriptors, one descriptor
         * per meta-property defined by the StpProperty. A MetaPropertyName
         * object is used as the property id.
         *
         * @return  An array of PropertyDescriptors for the meta-properties
         *          defined for this property.
         */
        public IPropertyDescriptor[] getPropertyDescriptors()
        {
            MetaPropertyName[] metaProps = m_property.metaPropertyNames();
            IPropertyDescriptor[] result =
                new IPropertyDescriptor[metaProps.length];

            for (int i = 0; i < result.length; ++i) {
                result[i] =
                    new PropertyDescriptor(metaProps[i], metaProps[i].getName());
            }

            return result;
        }

        /**
         * Retrieves the meta-property identified by a given property
         * identifier.
         *
         * @param  id  A MetaPropertyName identifying a meta-property of the
         *             StpProperty.
         *
         * @return  The value of the given meta-property, wrapped in a
         *          PropertySource if the item is a composite object such as a
         *          Resource, StpProperty, List, or Array.
         */
        public Object getPropertyValue(Object id)
        {
            try {
                MetaPropertyName<?> name = (MetaPropertyName<?>)id;
                Object prop = m_property.getMetaProperty(name);
                
                return getPropertySource(prop);
            } catch (WvcmException ex) {
                return ex;
            }
        }

        /** The StpProperty being displayed by this PropertySource */
        StpProperty<?> m_property;
    }

    /**
     * Selects a possible PropertySource for a given value based on the data
     * type of the value. A PropertySource is used for a value that is a
     * Resource, StpProperty, List (including ResourceList or StpProperty.List),
     * or an array.
     *
     * @param  val  The value to be displayed
     *
     * @return  The input value or an appropriate PropertySource for it.
     */
    private static Object getPropertySource(Object val)
    {
        if (val != null) {
            if (val instanceof List)
                val = new ListSource((List)val);
            else if (val instanceof StpResource)
                val = new ResourceSource((StpResource)val);
            else if (val.getClass().isArray())
                val = new ArraySource(val);
            else if (val instanceof StpProperty)
                val = new PropertySource((StpProperty)val);
            else if (val instanceof PropertyName) {
                PropertyName pn = (PropertyName)val;
                String       space = pn.getNamespace();

                val = pn.getName() + (space == null ? "" : " in " + space);
            }
        }

        return val;
    }

    /**
     * Constructs an appropriate display name for an item in a compound value.
     *
     * @param  item  The item to be displayed
     * @param  index  the index of the item in it's list.
     *
     * @return  A String containing a display name for the item.
     */
    private static String getDisplayName(
        Object item,
        int    index)
    {
        String name = "    " + Integer.toString(index);

        if (item instanceof StpResource)
            name = ((StpResource)item).location().lastSegment();
        else if (item instanceof StpProperty)
            name = ((StpProperty)item).getPropertyName().getName();
        else
            name = name.substring(name.length() - 5);

        return name;
    }
}
